/*
 * Copyright (c) 2014 by David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * An alternative easy to use curses package.
 * Initial termio support by Alan P.W. Hewett.
 */

#include <memory.h>
#include <signal.h>
#include <termcap.h>

#include "dpy_priv.h"


/*
 * Local routines.
 */
static	int dpySPutS(int);
static	void setStopHandler(void);
static	void dpyHandleStop(int arg);
static	void dpyFreeMemory(void);


/*
 * The window structure.
 */
WINDOW * _dpy_window_;


/*
 * Initialize the window structure.  Returns TRUE on success or FALSE on
 * failure with a message already typed.  The ttyType argument is the terminal
 * type string, or NULL if the TERM environment variable is to be used.
 * ModeString is a string whose characters describe the desired state of
 * the terminal.  These are:
 *	+	turn on mode indicated by next character (default)
 *	-	turn off mode indicated by next character.
 *	c	cbreak mode (return each character as typed, without waiting
 *		for an end of line, and leave signal processing on).
 *	e	echoing of typed-in characters is enabled.
 *	r	raw mode (return each character as typed, no canonical or
 *		signal processing of any kind).
 *	<SP>	spaces are ignored
 * So the terminal modes before dpy runs are normally "e -c -r".
 * A NULL modeString pointer defaults the modes to "-e c".
 */
BOOL
dpyInit(const char * ttyType, const char * modeString)
{
	WINDOW * wp;	/* window pointer */
	char * sp;	/* pointer for spaces */
	int size;	/* size of character array */
	int isOn;	/* whether or not mode is on */
	int modeChar;

	/*
	 * Allocate the window structure and initialise critical fields.
	 */
	wp = (WINDOW *) malloc(sizeof(WINDOW));

	if (wp == 0)
	{
		fprintf(stderr, "dpyInit: Cannot allocate window memory\n");

		return FALSE;
	}

	wp->isStopSafe = FALSE;
	wp->isStopPending = FALSE;
	wp->isInited = FALSE;
	wp->isDidOutput = FALSE;
	wp->begData = 0;

	_dpy_window_ = wp;

	/*
	 * Get the terminal type and then get the terminal data.
	 */
	if (ttyType == 0)
	{
		ttyType = getenv("TERM");

		if (ttyType == 0)
		{
			fprintf(stderr, "dpyInit: TERM not defined\n");

			dpyFreeMemory();

			return FALSE;
		}
	}

	size = tgetent(wp->tdata, ttyType);

	if (size <= 0)
	{
		fprintf(stderr, size ? "dpyInit: cannot open termcap file\n"
			: "dpyInit: unknown terminal type %s\n", ttyType);

		dpyFreeMemory();

		return FALSE;
	}

	/*
	 * Collect current tty modes, and remember editing characters
	 * and the output speed.
	 */
	dpyGetModes();

	/*
	 * Collect terminal capability strings
	 */
	wp->rowCount = tgetnum("li");
	wp->colCount = tgetnum("co");
	wp->tc_am = tgetflag("am");
	wp->tbuf[0] = '\0';
	wp->tptr = wp->tbuf;
/*	(void) tgetstr("pc", &wp->tptr);
	PC = wp->tbuf[0];*/
	wp->tptr = wp->tbuf;
	wp->tc_ho = tgetstr("ho", &wp->tptr);
	wp->tc_hocc = wp->tptr - wp->tbuf - 1;
	wp->tc_ce = tgetstr("ce", &wp->tptr);
	wp->tc_cd = tgetstr("cd", &wp->tptr);
	wp->tc_cm = tgetstr("cm", &wp->tptr);

	if ((wp->rowCount <= 0) || (wp->colCount <= 0) || (wp->tc_ce == 0)
		|| (wp->tc_cd == 0) || (wp->tc_cm == 0))
	{
		fprintf(stderr, "dpyInit: missing termcap entry\n");

		dpyFreeMemory();

		return FALSE;
	}

	/*
	 * Apply padding to clear screen and clear line.
	 */
	sp = wp->tptr;
	tputs(wp->tc_cd, wp->rowCount, dpySPutS);
	wp->tc_cd = sp;
	wp->tc_cdcc = wp->tptr - sp;

	sp = wp->tptr;
	tputs(wp->tc_ce, 1, dpySPutS);
	wp->tc_ce = sp;
	wp->tc_cecc = wp->tptr - sp;

	/*
	 * Make home up string if not defined.
	 */
	if ((wp->tc_ho == 0) || (wp->tc_hocc <= 0))
	{
		sp = wp->tptr;
		tputs(tgoto(wp->tc_cm, 0, 0), 1, dpySPutS);
		wp->tc_ho = sp;
		wp->tc_hocc = wp->tptr - sp;
	}

	/*
	 * Allocate and initialise the memory for the screen.
	 */
	if (!dpyAllocScreen(wp->rowCount, wp->colCount))
	{
		fprintf(stderr, "dpyInit: failed to allocate screen\n");

		dpyFreeMemory();

		return FALSE;
	}

	/*
	 * Set the default flags and tab size which will also set the
	 * effective ones for the initial full window.
	 */
	dpySetDefaultFlags(0);
	dpySetDefaultTabSize(DPY_INITIAL_TAB_SIZE);

	/*
	 * Modify the terminal modes as specified.
	 */
	wp->isRaw = FALSE;
	wp->isFull = FALSE;

	if (modeString == 0)
		modeString = "-e c";

	isOn = TRUE;

	while (*modeString)
	{
		modeChar = *modeString++;

		switch (modeChar)
		{
		case ' ':			/* spaces (ignored) */
			continue;

		case '+':			/* turn on next mode */
			isOn = TRUE;
			continue;

		case '-':			/* turn off next mode */
			isOn = FALSE;
			continue;

		case 'e':			/* enable echoing */
			dpyEcho(isOn);
			break;

		case 'c':			/* enable character mode */
			dpyCbreak(isOn);
			break;

		case 'r':			/* enable raw mode */
			dpyRaw(isOn);
			wp->isRaw = isOn;
			break;

		default:
			fprintf(stderr, "dpyInit: illegal flag: %c%c\n",
				(isOn ? '+' : '-'), modeChar);
	
			dpyFreeMemory();

			return FALSE;
		}

		isOn = TRUE;			/* reset mode */
	}

	wp->isInited = TRUE;

	/*
	 * Set the new terminal modes for real.
	 */
	dpySetModes(TRUE);

	/*
	 * If we have terminal stop capability, then enable that
	 * and set that stopping is safe now.
	 */
	setStopHandler();
	
	wp->isStopSafe = TRUE;

	return TRUE;
}


/*
 * Allocate new memory for the screen having the specified number of
 * rows and columns.  This also initialises the window structure elements
 * related to the window size and sets the current window to cover the
 * whole screen.  Returns TRUE on success, or FALSE on an error.
 */
BOOL
dpyAllocScreen(int rowCount, int colCount)
{
	WINDOW * wp = _dpy_window_;

	/*
	 * Calculate the size of the data required.
	 * This rounds up each line to a multiple of the integer size.
	 */
	int rowDelta = (colCount + INTSIZE) &~ (INTSIZE-1);
	int size = rowCount * rowDelta;

	/*
	 * Make sure the row and column counts are reasonable.
	 */
	if ((rowCount <= 0) || (rowCount > MAX_ROWCOUNT) ||
		(colCount <= 0) || (colCount > MAX_COLCOUNT))
	{
		return FALSE;
	}

	/*
	 * Allocate memory for two screen images and clear them to spaces.
	 */
	char * cp = malloc(2 * size);

	if (cp == 0)
		return FALSE;

	memset(cp, SPACE, 2 * size);

	/*
	 * Set up the data using the memory address.
	 */
	wp->rowCount = rowCount;
	wp->colCount = colCount;
	wp->rowDelta = rowDelta;
	wp->begData = cp;
	wp->endData = cp + size;
	wp->begWin = cp;
	wp->endWin = cp + size - rowDelta;
	wp->begRow = cp;
	wp->endRow = cp + colCount;
	wp->cp = cp;
	wp->screen = cp + size;
	wp->begChange = wp->endData;
	wp->endChange = wp->begData;
	wp->tabColumn = 0;
	wp->termRow = 0;
	wp->termCol = 0;
	wp->isFull = FALSE;

	return TRUE;
}


/*
 * Terminate the window, home down to the bottom of the screen, resets
 * the terminal modes to their original state, and resets our signals.
 * This does no terminal output if we haven't initialised and updated yet
 * and so can be called at any time.
 */
void
dpyClose(void)
{
	WINDOW * wp = _dpy_window_;

	if (wp == 0)
		return;

	wp->isStopSafe = FALSE;

	signal(SIGTSTP, SIG_DFL);
	signal(SIGWINCH, SIG_DFL);

	if (wp->isInited)
	{
		wp->isInited = FALSE;

		dpyClosePipeline();

		if (wp->isDidOutput)
		{
			dpyDoMove(wp->rowCount - 1, 0, 0);
			fwrite(wp->tc_ce, 1, wp->tc_cecc, stdout);
			fflush(stdout);
		}

		dpySetModes(FALSE);
	}

	dpyFreeMemory();
}


/*
 * Routine that should be called in the child process when a fork is done.
 * This frees the pipeline and resets our signal handlers.
 */
void
dpyChildClose(void)
{
	signal(SIGTSTP, SIG_DFL);
	signal(SIGWINCH, SIG_DFL);

	dpyClosePipeline();
}


/*
 * Set up to catch terminal stop signals if we can.
 */
static void
setStopHandler(void)
{
	WINDOW * wp = _dpy_window_;
	struct sigaction action;

	wp->pid = getpid();

	memset(&action, 0, sizeof(action));

	action.sa_handler = dpyHandleStop;

	(void) sigaction(SIGTSTP, &action, 0);
}


/*
 * Routine called on a terminal stop signal.
 * We stop ourself right now if it is safe, otherwise we just
 * set a flag to stop later when it is safe.
 */
static void
dpyHandleStop(int arg)
{
	WINDOW * wp = _dpy_window_;

	if ((wp == 0) || (wp->pid != getpid()))
		return;

	wp->isStopPending = TRUE;

	if (wp->isStopSafe)
		dpyStop();
}


/*
 * Routine called to stop dpy and allow it to continue.
 * Restore the original terminal state, home down to the bottom,
 * and really stop.  If continued, restore the new terminal state
 * and redraw the screen.  This clears the stop pending flag.
 * Returns TRUE if successful, or FALSE if stopping is not possible.
 */
BOOL
dpyStop(void)
{
	WINDOW * wp = _dpy_window_;
	BOOL isSuccess = TRUE;

	if ((wp == 0) || !wp->isInited)
		return FALSE;

	wp->isStopSafe = FALSE;
	wp->isStopPending = FALSE;

	if (wp->isDidOutput)
	{
		dpyDoMove(wp->rowCount - 1, 0, 0);
		fflush(stdout);
	}

	dpySetModes(FALSE);

	/*
	 * Now really stop.  If the user continues us then we will proceed.
	 * This stops every process in our process group.
	 */
	if (kill(0, SIGSTOP) < 0)
		isSuccess = FALSE;

	dpySetModes(TRUE);

	if (wp->isDidOutput)
		dpyRedraw();

	wp->isStopSafe = TRUE;

	return isSuccess;
}


/*
 * Local routine called by tputs to save a character.
 */
static int
dpySPutS(int ch)
{
	*_dpy_window_->tptr++ = ch;

	return 0;
}


/*
 * Free memory allocated by us if necessary.
 * This is the window structure and the screen memory.
 */
static void
dpyFreeMemory(void)
{
	WINDOW * wp = _dpy_window_;

	if (wp == 0)
		return;

	_dpy_window_ = 0;

	free(wp->begData);
	free(wp);
}

/* END CODE */
