/*
 * Copyright (c) 2014 by David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Public prototypes for the dpy library.
 */

#ifndef	DPY_H
#define	DPY_H

/*
 * The version of the dpy library.
 */
#define	DPY_VERSION	"4.0"

/*
 * Special return values from dpyInputChar, dpyPollInput, and dpyRead.
 * These values are all negative to indicate an input error.
 * The DPY_USER value (and ones further negative) are for the program's
 * own use.
 */
#define	DPY_EOL		-1	/* read end of line character */
#define	DPY_EOF		-2	/* read end of file condition */
#define	DPY_INTR	-3	/* read was interrupted */
#define	DPY_ERROR	-4	/* read got another error */
#define	DPY_USER	-10	/* user-defined error from dpyRead's routine */


/*
 * The flags that can be modified by dpySetFlags or dpySetDefaultFlags.
 * The default flags are reimposed after every dpyWindow or dpyFrame call.
 * The initial set of flags are all zero.
 */
#define	DPY_AUTO_SCROLL		0x0001	/* automatic scrolling is performed */
#define	DPY_LEAVE_CURSOR	0x0002	/* cursor is left after update */
#define	DPY_NO_LINE_WRAP	0x0004	/* line wrapping is not done */
#define	DPY_NO_TAB_ACTION	0x0008	/* tab action is disabled */
#define	DPY_NO_NEWLINE_ACTION	0x0010	/* newline action is disabled */
#define	DPY_NO_CTRLS_SHOWN	0x0020	/* control chars not shown as ^X */
#define	DPY_NO_HIGH_CHARS_SHOWN	0x0040	/* high bit chars not shown at all */
#define DPY_NO_HIGH_CHARS_META	0x0080	/* high bit chars not shown as M-x */

#define	DPY_ALL_FLAGS		0x00ff	/* all implemented flags */


/*
 * The size of the buffer used for dpyPrintf.
 * This allows a program to determine whether or not overflow might happen.
 */
#define	DPY_PRINTF_SIZE	5000


/*
 * The maximum tab size that can be set and the initial tab size.
 */
#define	DPY_MAX_TAB_SIZE	1000
#define	DPY_INITIAL_TAB_SIZE	8


/*
 * A boolean type used by some of the dpy routines.
 * We use our own definition since we don't want to conflict with other
 * booleans defined by the program.  A 0 is FALSE and nonzero (1) is TRUE.
 */
typedef	int DPY_BOOL;


/*
 * The dpy functions themselves.
 * These ones return a boolean value to indicate success or failure.
 * That is, 1 indicates success and 0 indicates failure.
 */
extern DPY_BOOL dpyInit(const char * ttyType, const char * modeString);
extern DPY_BOOL dpyWindow(int minRow, int maxRow, int minCol, int maxCol);
extern DPY_BOOL dpyFrame(int minRow, int maxRow, int minCol, int maxCol);
extern DPY_BOOL dpyMove(int row, int col);
extern DPY_BOOL dpyPlaceChar(int row, int col, int ch);
extern DPY_BOOL dpyWrite(const char * buf, int count);
extern DPY_BOOL dpyPrintf(const char * fmt, ...);
extern DPY_BOOL dpyChar(int ch);
extern DPY_BOOL dpyStr(const char * str);
extern DPY_BOOL dpyStop(void);
extern DPY_BOOL dpySave(char ** ptr);
extern DPY_BOOL dpyRestore(const char * ptr);
extern DPY_BOOL dpyGetSavedSize(const char *, int *, int *);
extern DPY_BOOL dpyRegisterResize(void (*func)(void));
extern DPY_BOOL dpySetInterrupt(void);
extern DPY_BOOL dpyClearInterrupt(void);


/*
 * These functions return an integer value which is zero or larger for
 * success and a negative value (typically -1) on failure.
 */
extern int dpyRead(const char *, int (*func)(int), char *, int, int);
extern int dpyCheckResize(void);
extern int dpyResize(void);
extern int dpyTestInput(void);
extern int dpyPollInput(int);
extern int dpyGetPollFileHandle(void);
extern int dpyInputChar(void);
extern int dpyGetRow(void);
extern int dpyGetCol(void);
extern int dpyGetChar(int row, int col);
extern int dpyGetSavedChar(const char *, int, int);
extern int dpySetDefaultFlags(int flags);
extern int dpySetFlags(int clearFlags, int setFlags);
extern int dpySetDefaultTabSize(int size);
extern int dpySetTabSize(int size);


/*
 * These procedures have no return value and always succeed.
 */
extern void dpyClearLine(void);
extern void dpyClearWindow(void);
extern void dpyHome(void);
extern void dpyUpdate(void);
extern void dpyRedraw(void);
extern void dpyClose(void);
extern void dpyChildClose(void);


/*
 * The original dpy functions.
 * These are under a conditional so that programs or Makefiles
 * have to be modified at least slightly to access them anymore.
 * This is an incentive to change them to use the newer calls above.
 * Besides the name changes, the return values are different.
 * In particular, the functions which return a boolean-like value
 * mostly use 0 for success and -1 on failure.  Also, there is no
 * support for terminal resizing.
 */

#ifdef DPY_USE_OLD_FUNCTIONS
extern int dpyinit(const char * ttyType, const char * modeString);
extern int dpywindow(int minRow, int maxRow, int minCol, int maxCol);
extern int dpyread(const char *, int (*routine)(int), char *, int, int);
extern int dpytestinput(void);
extern int dpyinputchar(void);
extern int dpymove(int row, int col);
extern int dpyplace(int row, int col, int ch);
extern int dpywrite(const char * buf, int count);
extern int dpyprintf(const char * fmt, ...);
extern int dpychar(int ch);
extern int dpystr(const char * str);
extern int dpyget(int row, int col);
extern void dpyclrline(void);
extern void dpyclrwindow(void);
extern void dpyhome(void);
extern int dpygetrow(void);
extern int dpygetcol(void);
extern void dpyupdate(void);
extern void dpyredraw(void);
extern int dpystop(void);
extern int dpysave(char ** ptr);
extern int dpyrestore(const char * ptr);
extern void dpyclose(void);

#define	DPYEOF		DPY_EOF
#define	DPYINTR		DPY_INTR
#define	DPYERROR	DPY_ERROR
#define	DPYVERSION	DPY_VERSION

#endif

#endif
