/*
 * Copyright (c) 2014 by David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */

#include <memory.h>

#include "dpy_priv.h"


/*
 * Structure which holds saved screen data.
 */
typedef struct
{
	long magic;		/* magic number */
	int rowCount;		/* number of rows of saved data */
	int colCount;		/* number of columns of saved data */
	int rowDelta;		/* delta between rows */
	char data[1];		/* the data (variable size - must be last) */
}
SAVEDATA;

#define	SAVE_MAGIC	1067485617L
#define	SAVE_SIZE(charCount)	(sizeof(SAVEDATA) + charCount)


static BOOL dpyIsValidSaveData(const SAVEDATA *);


/*
 * Save the current screen image for later restoration.
 * If the specified pointer is NULL, then we allocate enough memory for the
 * image and return the pointer.  If non-NULL, we try to reuse the memory
 * specified by the pointer, freeing it and allocating a new one if the
 * size differs from the terminal's current size.  This function does not
 * affect the current write position or the current window definition.
 * The image that is saved is the future screen image, not the actual
 * terminal screen image.  Returns TRUE on success or FALSE if the memory
 * could not be allocated or the old pointer was invalid.
 */
BOOL
dpySave(char ** ptr)
{
	WINDOW * wp = _dpy_window_;

	int charCount = wp->endData - wp->begData;
	int saveSize = SAVE_SIZE(charCount);
	SAVEDATA * saveData;

	/*
	 * Make sure we were given an address of a data pointer,
	 * and then retrieve it for examination.
	 */
	if (ptr == 0)
		return FALSE;

	saveData = (SAVEDATA *) *ptr;

	/*
	 * If an old pointer is supplied then check it for correctness.
	 * If the size doesn't match our current screen size then free
	 * the old data pointer so that we can allocate a new one.
	 */
	if (saveData)
	{
		if (!dpyIsValidSaveData(saveData))
			return FALSE;

		if ((saveData->rowCount != wp->rowCount) ||
			(saveData->colCount != wp->colCount))
		{
			*ptr = 0;
			free(saveData);
			saveData = 0;
		}
	}

	/*
	 * Now if we don't have a data pointer then allocate one.
	 */
	if (saveData == 0)
	{
		saveData = (SAVEDATA *) malloc(saveSize);

		if (saveData == 0)
			return FALSE;

		saveData->magic = SAVE_MAGIC;
		saveData->rowCount = wp->rowCount;
		saveData->colCount = wp->colCount;
		saveData->rowDelta = wp->rowDelta;
	}

	/*
	 * The structure is valid and has the correct dimensions.
	 * Copy the screen data into it and return the pointer.
	 */
	memcpy(saveData->data, wp->begData, charCount);

	*ptr = (char *) saveData;

	return TRUE;
}


/*
 * Restore a screen image that had been previously saved.
 * This does not affect the current write location or window definition.
 * The image that is restored is the future screen image, not the actual
 * terminal screen image.  The saved data can be a different size then the
 * terminal's current size, in which case only the common area is restored
 * with any leftover area set to spaces.  The supplied memory is not
 * freed or modified by us, so that the same pointer can be reused for
 * further save or restore calls, or it can be freed by the caller.
 * Returns TRUE on success or FALSE if the supplied pointer was invalid.
 */
BOOL
dpyRestore(const char * cp)
{
	WINDOW * wp = _dpy_window_;
	const SAVEDATA * saveData = (SAVEDATA *) cp;

	if (!dpyIsValidSaveData(saveData))
		return FALSE;

	dpyCopyData(saveData->data, saveData->rowCount, saveData->colCount,
		saveData->rowDelta, wp->begData, wp->rowCount, wp->colCount,
		wp->rowDelta);

	/*
	 * Mark that the whole screen is possibly changed.
	 */
	wp->begChange = wp->begData;
	wp->endChange = wp->endData;

	return TRUE;
}


/*
 * Return the character at the specified row and column from the saved data.
 * Returns -1 if the data pointer is invalid or if the row or column
 * numbers are out of range.
 */
int
dpyGetSavedChar(const char * cp, int row, int col)
{
	const SAVEDATA * saveData = (SAVEDATA *) cp;

	if (!dpyIsValidSaveData(saveData))
		return -1;

	if (((unsigned) row >= saveData->rowCount) ||
		((unsigned) col >= saveData->colCount))
	{
		return -1;
	}

	return (saveData->data[row * saveData->rowDelta + col] & 0xff);
}


/*
 * Return the number of rows and columns in the specified saved data.
 * Returns TRUE on success, or FALSE if the supplied pointer is invalid.
 */
BOOL
dpyGetSavedSize(const char * cp, int * retRowCount, int * retColCount)
{
	const SAVEDATA * saveData = (SAVEDATA *) cp;

	*retRowCount = 0;
	*retColCount = 0;

	if (!dpyIsValidSaveData(saveData))
		return FALSE;

	*retRowCount = saveData->rowCount;
	*retColCount = saveData->colCount;

	return TRUE;
}


/*
 * Return the size of the data that has been allocated in the save.
 * Returns the size on success, or -1 if the supplied pointer is invalid.
 * This doesn't have much utility, and is only defined privately so that
 * it can be used by the old code.
 */
int
dpyGetSavedAllocation(const char * cp)
{
	const SAVEDATA * saveData = (SAVEDATA *) cp;

	if (!dpyIsValidSaveData(saveData))
		return -1;

	return SAVE_SIZE(saveData->rowCount * saveData->rowDelta);
}


/*
 * Check the specifed save data structure to see if it is valid.
 * Returns TRUE if the data seems valid.
 */
static BOOL
dpyIsValidSaveData(const SAVEDATA * saveData)
{
	if ((saveData == 0) || (saveData->magic != SAVE_MAGIC) ||
		(saveData->rowCount <= 0) ||
		(saveData->rowCount > MAX_ROWCOUNT) ||
		(saveData->colCount <= 0) ||
		(saveData->colCount > MAX_COLCOUNT) ||
		(saveData->rowDelta <= 0))
	{
		return FALSE;
	}

	return TRUE;
}

/* END CODE */
